from typing import List
from .BasicDXF import BasicDXF
from .DrawingAttribs import (
    DrawingAttribs,
    PolylineAttribs,
    TextAttribs,
    TextEntityAlignment,
)
from .LayerManager import DefaultLayers, CADColor
from CivilTools.Const import Concrete, ConcreteLevel, Steel, SteelLevel


class DetailDXF(BasicDXF):
    def __init__(self):
        super().__init__()


default_floor_height_table_attribs = {
    "column_con_level": None,
    "column_steel_level": None,
    "beam_con_level": None,
    "beam_steel_level": None,
}


class FloorInformation:
    def __init__(
        self,
        name: str,
        archi_elevation: float,
        plaster_thickness: float = 200,
        beam_slab_con_mat: ConcreteLevel = Concrete.C35,
        wall_con_mat: ConcreteLevel = Concrete.C60,
        column_con_mat: ConcreteLevel = Concrete.C50,
        brace_con_mat: ConcreteLevel = Concrete.C50,
        beam_steel_mat: SteelLevel = Steel.HRB300,
        column_steel_mat: SteelLevel = Steel.Q355,
        brace_steel_mat: SteelLevel = Steel.Q355,
    ):
        if plaster_thickness < 0:
            raise ValueError("Plaster value should be non-negitive.")
        self.floor_name = name
        self.__struct_elevation = archi_elevation - plaster_thickness
        self.__archi_elevation = archi_elevation

        self.beam_slab_con_mat = beam_slab_con_mat
        self.wall_con_mat = wall_con_mat
        self.column_con_mat = column_con_mat
        self.brace_con_mat = brace_con_mat

        self.beam_steel_mat = beam_steel_mat
        self.column_steel_mat = column_steel_mat
        self.brace_steel_mat = brace_steel_mat
        self.is_embedded = False

    def get_attribute_by_name(self, attrib_name: str):
        if "楼层号" in attrib_name:
            return self.floor_name
        if "建筑" in attrib_name and "标高" in attrib_name:
            return self.archi_elevation
        if "结构" in attrib_name and "标高" in attrib_name:
            return self.struct_elevation
        if "柱" in attrib_name:
            return self.column_con_mat.name
        if "墙" in attrib_name:
            return self.wall_con_mat.name
        if "板" in attrib_name:
            return self.beam_slab_con_mat.name

    @property
    def struct_elevation(self):
        if self.__struct_elevation == 0:
            return "±0.000"
        return f"{self.__struct_elevation/1000:.3f}"

    @property
    def archi_elevation(self):
        if self.__archi_elevation == 0:
            return "±0.000"
        return f"{self.__archi_elevation/1000:.3f}"


class FloorHeightTableDXF(BasicDXF):
    def __prepare(self):
        useful_layers = [DefaultLayers.TEXT, DefaultLayers.SYMBOL]
        self.init_layers(useful_layers)

    def __init__(
        self,
        above_floor_num: int,
        under_floor_num: int,
        base_elevation: float,
        font_size: float = 300,
    ):
        super().__init__()
        self.__prepare()
        self.__above_floor_num = above_floor_num
        self.__under_floor_num = under_floor_num
        self.__base_elevation = base_elevation
        self.floor_num = above_floor_num + under_floor_num
        self.floors = self.__create_floors()
        self.__font_size = font_size
        self.__table_title = None
        self.has_embeded_floor = False
        self.embeded_floor = None
        self.insert_point = [0, 0]

        for key, value in default_floor_height_table_attribs.items():
            setattr(self, key, value)

    def __create_floors(self) -> List[FloorInformation]:
        result = []
        for i in range(self.__above_floor_num):
            temp_floor = FloorInformation(f"F{i+1}", self.__base_elevation + 3500 * i)
            result.append(temp_floor)
        for i in range(self.__under_floor_num):
            temp_floor = FloorInformation(
                f"B{i+1}", self.__base_elevation - 3500 * (i + 1)
            )
            result.insert(0, temp_floor)
        return result

    def export_dxf(self, path):
        self._remove_all_entities()
        self.__draw_table_title()
        self.__analysis_table_columns()
        self.__draw_table_grid()
        self.__draw_table_column_title()
        self.__draw_context()
        self.__draw_supplement_info()
        self._save(path)

    def set_table_title(self, title: str):
        self.__table_title = title

    def set_font_size(self, font_size: float):
        self.__font_size = font_size

    def set_embeding_floor(self, floor_name: str | None):
        """Use None to reset embeded floor."""
        if floor_name == None:
            self.has_embeded_floor = False
            self.embeded_floor = None
            return
        for floor in self.floors:
            if floor.floor_name == floor_name:
                floor.is_embedded = True
                self.has_embeded_floor = True
                self.embeded_floor = floor
                return
        raise ValueError(f"No floor named as {floor_name}")

    def __analysis_table_columns(self):
        self.__columns = []
        self.__column_widths = []

        self.__columns.append("当前层")
        self.__column_widths.append(3)

        self.__columns.append("建筑楼层号")
        self.__column_widths.append(5)

        self.__columns.append("建筑标高(m)")
        self.__column_widths.append(6)

        self.__columns.append("结构标高(m)")
        self.__column_widths.append(6)

        self.__first_part_column_num = len(self.__columns)

        self.__columns.append("结构柱")
        self.__column_widths.append(3)

        self.__columns.append("剪力墙")
        self.__column_widths.append(3)

        self.__columns.append("梁/板")
        self.__column_widths.append(3)

    def __draw_table_title(self):
        if self.__table_title == None:
            return
        text_attrib = TextAttribs(
            DefaultLayers.TEXT.name,
            text_height=self.__font_size,
            text_align=TextEntityAlignment.MIDDLE_LEFT,
        )
        self._add_text(
            self.__table_title,
            [self.insert_point[0], self.insert_point[1] + self.__font_size * 1.5],
            text_attrib,
        )

    def __draw_table_grid(self):
        first_part_column_num = self.__first_part_column_num
        start_x = self.insert_point[0]
        start_y = self.insert_point[1]
        font_factor = self.__font_size

        total_width = font_factor * sum(self.__column_widths)
        first_part_width = font_factor * sum(
            self.__column_widths[:first_part_column_num]
        )

        total_height = font_factor * 2 * (self.floor_num + 2)
        grid_draw_attrib = PolylineAttribs(DefaultLayers.SYMBOL.name)
        for i in range(self.floor_num + 3):
            if i == 1:
                self._add_horizental_line(
                    start_x + first_part_width,
                    start_y,
                    total_width - first_part_width,
                    grid_draw_attrib,
                )
            else:
                self._add_horizental_line(
                    start_x, start_y, total_width, grid_draw_attrib
                )
            start_y -= font_factor * 2
            if (
                i - 2 >= 0
                and i - 2 < self.floor_num
                and self.floors[::-1][i - 2].is_embedded
            ):
                self._add_horizental_line(
                    start_x, start_y + font_factor * 0.2, total_width, grid_draw_attrib
                )
        start_y = 0
        for i in range(len(self.__columns) + 1):
            if i <= first_part_column_num or i == len(self.__columns):
                self._add_vertical_line(
                    start_x, start_y, -total_height, grid_draw_attrib
                )
            else:
                self._add_vertical_line(
                    start_x,
                    start_y - font_factor * 2,
                    -total_height + font_factor * 2,
                    grid_draw_attrib,
                )
            if i == len(self.__columns):
                continue
            start_x += font_factor * self.__column_widths[i]

    def __draw_table_column_title(self):
        font_factor = self.__font_size
        start_x = self.insert_point[0]
        start_y = self.insert_point[1]

        table_column_title_attrib = TextAttribs(
            DefaultLayers.TEXT.name,
            text_height=font_factor,
            text_align=TextEntityAlignment.MIDDLE_CENTER,
        )

        for i in range(len(self.__columns)):
            temp_string = self.__columns[i]
            temp_x = (
                sum(self.__column_widths[: i + 1]) - self.__column_widths[i] / 2
            ) * font_factor
            temp_y = (
                -font_factor * 2
                if i < self.__first_part_column_num
                else -font_factor * 3
            )
            self._add_text(
                temp_string,
                [start_x + temp_x, start_y + temp_y],
                table_column_title_attrib,
            )
        temp_x = (
            (
                sum(self.__column_widths)
                + sum(self.__column_widths[: self.__first_part_column_num])
            )
            / 2.0
            * font_factor
        )
        temp_y = -font_factor * 1
        self._add_text("混凝土结构", [temp_x, temp_y], table_column_title_attrib)

    def __draw_context(self):
        font_factor = self.__font_size

        start_x = self.insert_point[0]
        start_y = self.insert_point[1] - font_factor * 4
        for i in range(len(self.floors)):
            temp_floor = self.floors[::-1][i]
            for j in range(1, len(self.__columns)):
                temp_insert_point = [
                    start_x + font_factor * (sum(self.__column_widths[: j + 1]) - 0.5),
                    start_y - font_factor * (2 * i + 1),
                ]
                temp_string = temp_floor.get_attribute_by_name(self.__columns[j])
                self._add_text(
                    temp_string,
                    temp_insert_point,
                    TextAttribs(
                        DefaultLayers.TEXT.name,
                        color_index=CADColor.Yellow.value,
                        text_height=font_factor,
                        text_align=TextEntityAlignment.MIDDLE_RIGHT,
                    ),
                )

    def __draw_supplement_info(self):
        if not self.has_embeded_floor:
            return
        font_factor = self.__font_size
        temp_x = self.insert_point[0]
        temp_y = self.insert_point[1] - font_factor * (self.floor_num + 2.5) * 2
        table_supplement_info_attrib = TextAttribs(
            DefaultLayers.TEXT.name,
            text_height=font_factor,
            text_align=TextEntityAlignment.MIDDLE_LEFT,
        )
        self._add_text(
            f"上部嵌固层：{self.embeded_floor.floor_name}，标高：{self.embeded_floor.struct_elevation}。",
            [temp_x, temp_y],
            table_supplement_info_attrib,
        )

    def __draw_embeding_dimension(self):
        return NotImplemented



from .ShapeEnum import ShapeEnum
from typing import List


class Section:
    def __init__(self, id: int, k: ShapeEnum, vals: List[float], mat: int = 1):
        self.id = id
        self.k = k
        self.mat = mat
        value_number = len(vals)
        self.b = vals[0] if value_number >= 1 else None
        self.h = vals[1] if value_number >= 2 else None
        self.u = vals[2] if value_number >= 3 else None
        self.t = vals[3] if value_number >= 4 else None
        self.d = vals[4] if value_number >= 5 else None
        self.f = vals[5] if value_number >= 6 else None

    def __str__(self):
        """todo:这里需要完善str方法"""
        display_function = {
            ShapeEnum.Rect: lambda: f"Rect-{self.b}mmx{self.h}mm",
            ShapeEnum.HShape: lambda: f"{self.b}",
            ShapeEnum.Circle: lambda: f"Circle-Diameter:{self.b}mm",
            ShapeEnum.RegularPolygon: lambda: f"{self.b}",
            ShapeEnum.Groove: lambda: f"{self.b}",
            ShapeEnum.Cross: lambda: f"{self.b}",
            ShapeEnum.Box: lambda: f"{self.b}",
            ShapeEnum.CircleTube: lambda: f"{self.b}",
            ShapeEnum.CircleCFT: lambda: f"{self.b}",
            ShapeEnum.HSRC: lambda: f"{self.b}",
            ShapeEnum.BoxSRC: lambda: f"{self.b}",
            ShapeEnum.CrossSRC: lambda: f"{self.b}",
            ShapeEnum.UnKnown: lambda: f"Unknown:{self.b}",
        }
        return display_function[self.k]()

    def __repr__(self):
        return self.__str__()


if __name__ == "__main__":
    s = Section(1, ShapeEnum.Circle, [20, 32])

    print(s)



from typing import List


class RowDataFactory:
    """
    This Class is used to extract data form row_data
    """

    @classmethod
    def extract_int(cls, row_data: List, index: int):
        RowDataFactory.__list_index_check(row_data, index)
        return RowDataFactory.convert_to_int(row_data[index])

    @classmethod
    def extract_float(cls, row_data: List, index: int):
        RowDataFactory.__list_index_check(row_data, index)
        return RowDataFactory.convert_to_float(row_data[index])

    @classmethod
    def extract_list(cls, row_data: List, index: int):
        RowDataFactory.__list_index_check(row_data, index)
        try:
            result = list(row_data[index].split(","))
        except SyntaxError:
            raise TypeError(f"Value: {row_data[index]} cannot be converted to list.")
        return result

    @classmethod
    def convert_to_float(cls, string: "str"):
        try:
            result = (float)(string)
        except SyntaxError:
            raise TypeError(f"Value: {string} cannot be converted to float.")
        return result

    @classmethod
    def convert_to_int(cls, string: "str"):
        try:
            result = (int)(string)
        except SyntaxError:
            raise TypeError(f"Value: {string} cannot be converted to int.")
        return result

    @classmethod
    def __list_index_check(cls, row_data: List, index: int):
        if len(row_data) <= index:
            raise IndexError("Index out of the range.")



from docx.oxml.xmlchemy import BaseOxmlElement
from docx.oxml.ns import qn, nsdecls
from CivilTools.YDBLoader.BuildingDefine.StairPart import StairPart, Component
import numpy as np


# 网上找的代码，有点东西
# https://stackoverflow.com/questions/33069697/how-to-setup-cell-borders-with-python-docx
# 这个是github上的讨论帖，在讨论帖中找到的
# https://github.com/python-openxml/python-docx/issues/1306
def set_cell_border(cell, **kwargs):
    """
    Set cell`s border
    Usage:

    set_cell_border(
        cell,
        top={"sz": 12, "val": "single", "color": "#FF0000", "space": "0"},
        bottom={"sz": 12, "color": "#00FF00", "val": "single"},
        start={"sz": 24, "val": "dashed", "shadow": "true"},
        end={"sz": 12, "val": "dashed"},
    )
    """
    tc = cell._tc
    tc_pr = tc.get_or_add_tcPr()

    # check for tag existnace, if none found, then create one
    tc_borders = tc_pr.first_child_found_in("w:tcBorders")
    if tc_borders is None:
        tc_borders = BaseOxmlElement("w:tcBorders")
        tc_pr.append(tc_borders)

    # list over all available tags
    for edge in ("start", "top", "end", "bottom", "insideH", "insideV"):
        edge_data = kwargs.get(edge)
        if edge_data:
            tag = "w:{}".format(edge)

            # check for tag existnace, if none found, then create one
            element = tc_borders.find(qn(tag))
            if element is None:
                element = BaseOxmlElement(tag)
                tc_borders.append(element)

            # looks like order of attributes is important
            for key in ["sz", "val", "color", "space", "shadow"]:
                if key in edge_data:
                    element.set(qn("w:{}".format(key)), str(edge_data[key]))


def analysis_sub_and_super_script(context: str):
    """根据字符串中的'_'与'^'标志的上下标（上下标需要被{}包围起来），将字符串分隔并返回上下标结果
    返回的sub_or_super列表中，0代表常规字符，1代表下标，2代表上标，3代表highlighted
    Args:
        context (str): 输入的文字
    """
    contexts = []
    sub_or_super = [0]
    i = 0
    j = 0
    length = len(context)
    flag = False
    index_for_flag = {"_": 1, "^": 2, "*": 3}

    for c in context:
        if (c in index_for_flag.keys()) and (j < length and context[j + 1] == "{"):
            flag = True
            contexts.append(context[i:j])
            sub_or_super.append(index_for_flag[c])
            i = j + 2
        if flag and c == "}":
            contexts.append(context[i:j])
            sub_or_super.append(0)
            i = j + 1
            flag = False
        j += 1
    contexts.append(context[i:j])
    return contexts, sub_or_super


def add_comma_in_num_str(num: int):
    if not isinstance(num, int):
        raise ValueError("Only int number can be added.")
    num_str = str(num)[::-1]
    result = ""
    for i in range(len(num_str)):
        if i > 0 and i % 3 == 0 and num_str[i] != "-":
            result = "," + result
        result = num_str[i] + result
    return result


class MatrixSolver:
    def __init__(self, stair: StairPart):
        self.stair = stair
        self.cut_point = 0.5

    def set_load(self, q1, q2, q3):
        self.q1 = q1
        self.q2 = q2
        self.q3 = q3

    def pre_analysis(self):
        concrete_e = 30000
        uniform_b = 1000
        top_y = (
            self.stair.position.higher_elevation - self.stair.position.lower_elevation
        )
        if self.stair.stair_type == "AT":
            self.point_num = 3
            self.boundary = np.array([0, 0, 1, 1, 1, 1, 0, 0, 1])

            point1 = [0, 0]
            point2 = [self.stair.total_horizental_length, top_y]
            point_middle = self.calculate_middle_point(point1, point2, self.cut_point)
            comp1 = Component(point1, point_middle)
            comp2 = Component(point_middle, point2)
            comp1.set_vertical_q(self.q2)
            comp1.set_comp_attr(concrete_e, uniform_b, self.stair.main_thick)
            comp2.set_vertical_q(self.q2)
            comp2.set_comp_attr(concrete_e, uniform_b, self.stair.main_thick)
            self.comp_list = [comp1, comp2]

        elif self.stair.stair_type == "BT":
            self.point_num = 4
            self.q_list = [self.q1, self.q2, self.q2]
            self.boundary = np.array([0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1])
            point0 = [0, 0]
            point1 = [self.stair.left_extend_length, 0]
            point2 = [self.stair.total_horizental_length, top_y]
            point_middle = self.calculate_middle_point(point1, point2, self.cut_point)
            comp1 = Component(point0, point1)
            comp2 = Component(point1, point_middle)
            comp3 = Component(point_middle, point2)
            comp1.set_vertical_q(self.q1)
            comp1.set_comp_attr(concrete_e, uniform_b, self.stair.left_thick)
            comp2.set_vertical_q(self.q2)
            comp2.set_comp_attr(concrete_e, uniform_b, self.stair.main_thick)
            comp3.set_vertical_q(self.q2)
            comp3.set_comp_attr(concrete_e, uniform_b, self.stair.main_thick)
            self.comp_list = [comp1, comp2, comp3]

        elif self.stair.stair_type == "CT":
            self.point_num = 4
            self.q_list = [self.q2, self.q2, self.q3]
            self.boundary = np.array([0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1])

            point1 = [0, 0]
            point2 = [
                self.stair.total_horizental_length - self.stair.right_extend_length,
                top_y,
            ]
            point3 = [self.stair.total_horizental_length, top_y]
            point_middle = self.calculate_middle_point(point1, point2, self.cut_point)
            comp1 = Component(point1, point_middle)
            comp2 = Component(point_middle, point2)
            comp3 = Component(point2, point3)
            comp1.set_vertical_q(self.q2)
            comp1.set_comp_attr(concrete_e, uniform_b, self.stair.main_thick)

            comp2.set_vertical_q(self.q2)
            comp2.set_comp_attr(concrete_e, uniform_b, self.stair.main_thick)

            comp3.set_vertical_q(self.q3)
            comp3.set_comp_attr(concrete_e, uniform_b, self.stair.right_thick)
            self.comp_list = [comp1, comp2, comp3]

        else:
            self.point_num = 5
            self.q_list = [self.q1, self.q2, self.q2, self.q3]
            self.boundary = np.array([0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1])

            point0 = [0, 0]
            point1 = [self.stair.left_extend_length, 0]
            point2 = [
                self.stair.total_horizental_length - self.stair.right_extend_length,
                top_y,
            ]
            point3 = [self.stair.total_horizental_length, top_y]
            point_middle = self.calculate_middle_point(point1, point2, self.cut_point)
            comp1 = Component(point0, point1)
            comp2 = Component(point1, point_middle)
            comp3 = Component(point_middle, point2)
            comp4 = Component(point2, point3)

            comp1.set_vertical_q(self.q1)
            comp1.set_comp_attr(concrete_e, uniform_b, self.stair.left_thick)
            comp2.set_vertical_q(self.q2)
            comp2.set_comp_attr(concrete_e, uniform_b, self.stair.main_thick)
            comp3.set_vertical_q(self.q2)
            comp3.set_comp_attr(concrete_e, uniform_b, self.stair.main_thick)
            comp4.set_vertical_q(self.q3)
            comp4.set_comp_attr(concrete_e, uniform_b, self.stair.right_thick)
            self.comp_list = [comp1, comp2, comp3, comp4]

    def calculate_middle_point(self, p1, p2, cut_point):
        x = p1[0] + (p2[0] - p1[0]) * cut_point
        y = p1[1] + (p2[1] - p1[1]) * cut_point
        return [x, y]

    def submit_problem(self):
        self.pre_analysis()
        comp_num = len(self.comp_list)
        total_k = np.zeros((comp_num * 3 + 3, comp_num * 3 + 3))
        total_f = np.zeros((comp_num * 3 + 3, 1))
        for i in range(len(self.comp_list)):
            comp = self.comp_list[i]
            temp_k = comp.create_k()
            temp_f = comp.create_f()
            total_k[i * 3 : i * 3 + 6, i * 3 : i * 3 + 6] += temp_k
            total_f[i * 3 : i * 3 + 6, :] += temp_f
        # 生成缩减矩阵并计算
        new_k = np.delete(total_k, np.nonzero(self.boundary == 0), axis=0)
        new_k = np.delete(new_k, np.nonzero(self.boundary == 0), axis=1)
        new_f = np.delete(total_f, np.nonzero(self.boundary == 0), axis=0)
        new_u = np.dot(np.linalg.inv(new_k), new_f)

        total_u = self.boundary.copy().reshape(-1, 1).astype(float)
        total_u[np.nonzero(total_u == 1)[0]] = new_u

        # 计算每个构件端点的数据
        for i in range(len(self.comp_list)):
            comp = self.comp_list[i]
            comp_u = total_u[i * 3 : i * 3 + 6]
            comp_f = np.dot(comp.create_k(), comp_u)
            comp_f -= comp.create_f()
            comp.set_f(comp_f[0][0], comp_f[1][0], comp_f[3][0], comp_f[4][0])
            comp.set_m(comp_f[2][0], comp_f[5][0])
        return self.comp_list
